import os
import re

from django import forms
from django.utils.translation import ugettext as _


alnum_name_re = re.compile(r'^[\sa-zA-Z0-9._/-]+$')


class FileForm(forms.Form):

    def __init__(self, path, *args, **kwargs):
        self.path = path
        super(FileForm, self).__init__(*args, **kwargs)

    def _normalize_name(self, name):
        return name.replace(' ', '_').lower()

    def _check_if_name_correct(self, name):
        if not alnum_name_re.search(name):
            raise forms.ValidationError(_(
                u'Only letters, numbers, underscores, spaces and hyphens are '
                u'allowed.'
            ))

    def _check_if_not_exists(self, name):
        if os.path.isdir(os.path.join(self.path, name)):
            raise forms.ValidationError(_(
                u'The folder with this name already exists.'
            ))
        if os.path.isfile(os.path.join(self.path, name)):
            raise forms.ValidationError(_(
                u'The file with this name already exists.'
            ))


class MakeDirForm(FileForm):

    dir_name = forms.CharField(
        label=_(u'Name'),
        min_length=3,
        max_length=50,
        help_text=_(
            u'Only letters, numbers, underscores, spaces and hyphens are '
            u'allowed.'
        ),
    )

    def clean_dir_name(self):
        dir_name = self.cleaned_data['dir_name']
        self._check_if_name_correct(dir_name)
        dir_name = self._normalize_name(dir_name)
        self._check_if_not_exists(dir_name)
        return dir_name


class RenameForm(FileForm):

    name = forms.CharField(
        label=_(u'New Name'),
        min_length=3,
        max_length=50,
        help_text=_(
            u'Only letters, numbers, underscores, spaces and hyphens are '
            u'allowed.'
        ),
    )

    def __init__(self, file_extension, *args, **kwargs):
        self.file_extension = file_extension
        super(RenameForm, self).__init__(*args, **kwargs)

    def clean_name(self):
        name = self.cleaned_data.get('name')
        self._check_if_name_correct(name)
        name_ext = self._normalize_name(name + self.file_extension)
        self._check_if_not_exists(name_ext)
        return name


class FileUploadForm(FileForm):

    file = forms.FileField(
        label=_(u'File'),
        help_text=_(
            u'Only letters, numbers, underscores, spaces and hyphens are '
            u'allowed.'
        ),
    )

    def clean_file(self):
        uploaded_file = self.cleaned_data.get('file')
        self._check_if_name_correct(uploaded_file.name)
        uploaded_file.name = self._normalize_name(uploaded_file.name)
        self._check_if_not_exists(uploaded_file.name)
        return uploaded_file



from django.dispatch import Signal


filebrowser_pre_createdir = Signal(providing_args=['path', 'dirname'])
filebrowser_post_createdir = Signal(providing_args=['path', 'dirname'])

filebrowser_pre_delete = Signal(providing_args=['path', 'filename'])
filebrowser_post_delete = Signal(providing_args=['path', 'filename'])

filebrowser_pre_rename = Signal(
    providing_args=['path', 'filename', 'new_filename'],
)
filebrowser_post_rename = Signal(
    providing_args=['path', 'filename', 'new_filename'],
)



import datetime
import os
import re

from django.utils import formats
from django.utils.text import capfirst

from .settings import MEDIA_ROOT, MEDIA_URL


class FileObject(object):

    def __init__(self, path, file_browser_site):
        self.path = path
        self.file_browser_site = file_browser_site
        self.head = os.path.split(path)[0]
        self.filename = os.path.split(path)[1]
        self.filename_lower = self.filename.lower() # important for sorting
        self.filetype = self._get_filetype()
    
    def _get_filetype(self):
        file_extension = os.path.splitext(self.filename_lower)[1]
        if os.path.isdir(os.path.join(MEDIA_ROOT, self.path)):
            return 'Folder'
        for key, value in self.file_browser_site.extensions.iteritems():
            for extension in value:
                if file_extension == extension.lower():
                    return key
        return ''

    def _filesize(self):
        if os.path.isfile(os.path.join(MEDIA_ROOT, self.path)) or \
           os.path.isdir(os.path.join(MEDIA_ROOT, self.path)):
            return os.path.getsize(os.path.join(MEDIA_ROOT, self.path))
        return ""
    filesize = property(_filesize)

    def _date(self):
        if os.path.isfile(os.path.join(MEDIA_ROOT, self.path)) or \
           os.path.isdir(os.path.join(MEDIA_ROOT, self.path)):
            return os.path.getmtime(os.path.join(MEDIA_ROOT, self.path))
        return ""
    date = property(_date)

    def _datetime(self):
        return datetime.datetime.fromtimestamp(self.date)
    datetime = property(_datetime)

    def _formatted_datetime(self):
        return capfirst(formats.date_format(self.datetime, 'DATETIME_FORMAT'))
    formatted_datetime = property(_formatted_datetime)

    def _extension(self):
        return u"%s" % os.path.splitext(self.filename)[1]
    extension = property(_extension)

    def _filetype_checked(self):
        if self.filetype == "Folder" and os.path.isdir(self.path_full):
            return self.filetype
        elif self.filetype != "Folder" and os.path.isfile(self.path_full):
            return self.filetype
        else:
            return ""
    filetype_checked = property(_filetype_checked)

    def _path_full(self):
        return u"%s" % os.path.join(MEDIA_ROOT, self.path)
    path_full = property(_path_full)

    def _path_relative_directory(self):
        directory_re = re.compile(r'^(%s)' % self.file_browser_site.upload_to)
        value = directory_re.sub('', self.path).lstrip('/')
        return u"%s" % value
    path_relative_directory = property(_path_relative_directory)

    def _url_full(self):
        return u"%s" % self._url_join(MEDIA_URL, self.path)
    url_full = property(_url_full)

    def _is_empty(self):
        if os.path.isdir(self.path_full):
            if not os.listdir(self.path_full):
                return True
            else:
                return False
        else:
            return None
    is_empty = property(_is_empty)
    
    def _url_join(self, *args):
        if args[0].startswith("http://"):
            url = "http://"
        else:
            url = "/"
        for arg in args:
            arg = unicode(arg).replace("\\", "/")
            arg_split = arg.split("/")
            for elem in arg_split:
                if elem != "" and elem != "http:":
                    url = url + elem + "/"
        # remove trailing slash for filenames
        if os.path.splitext(args[-1])[1]:
            url = url.rstrip("/")
        return url

    def __repr__(self):
        return u'%s' % self.path

    def __str__(self):
        return '%s' % self.path

    def __unicode__(self):
        return u'%s' % self.path



from django.utils.safestring import mark_safe


def query_helper(query, *remove, **add):
    return get_query_string(query.copy(), remove, add)


def get_query_string(p, remove=None, add=None):
    if add is None:
        add = {}
    if remove is None:
        remove = []
    for r in remove:
        for k in p.keys():
            if k == r:
                del p[k]
    for k, v in add.items():
        if k in p and v is None:
            del p[k]
        elif v is not None:
            p[k] = v
    params = ['%s=%s' % (k, v) for k, v in p.items()]
    return mark_safe('?' + '&'.join(params).replace(' ', '%20'))


def string_to_dict(string):
    kwargs = {}
    if string:
        string = str(string)
        if ',' not in string:
            # ensure at least one ','
            string += ','
        for arg in string.split(','):
            arg = arg.strip()
            if arg == '':
                continue
            kw, val = arg.split('=', 1)
            kwargs[kw] = val
    return kwargs


def string_to_list(string):
    args = []
    if string:
        string = str(string)
        if ',' not in string:
            # ensure at least one ','
            string += ','
        for arg in string.split(','):
            arg = arg.strip()
            if arg == '':
                continue
            args.append(arg)
    return args



import re

from django import template
from django.template.loader import render_to_string

from ..utils import get_query_string, string_to_list, string_to_dict


DOT = '.'

register = template.Library()


class QueryStringNode(template.Node):

    def __init__(self, remove=None, add=None):
        self.add = string_to_dict(add)
        self.remove = string_to_list(remove)

    def render(self, context):
        query = context.get('query')
        if not query:
            return ''
        return get_query_string(query.copy(), self.remove, self.add)


@register.tag
def query_string(parser, token):
    bits = token.split_contents()
    tag_name = bits[0]
    if len(bits) > 1:
        remove = bits[1]
        if not (remove[0] == remove[-1] and remove[0] in ('"', "'")):
            raise template.TemplateSyntaxError(
                u"%r tag's arguments should be in quotes" % tag_name)
        if len(bits) > 2:
            add = bits[2]
            if not (add[0] == add[-1] and add[0] in ('"', "'")):
                raise template.TemplateSyntaxError(
                    u"%r tag's arguments should be in quotes" % tag_name)
            return QueryStringNode(remove[1:-1], add[1:-1])
        return QueryStringNode(remove[1:-1])
    return QueryStringNode()


@register.inclusion_tag('filebrowser/tags/paginator.html', takes_context=True)
def pagination(context):
    page = context['page']
    page_num = page.number - 1
    paginator = page.paginator

    if not paginator.num_pages or paginator.num_pages == 1:
        page_range = []
    else:
        ON_EACH_SIDE = 3
        ON_ENDS = 2
        if paginator.num_pages <= 10:
            page_range = range(paginator.num_pages)
        else:
            page_range = []
            if page_num > (ON_EACH_SIDE + ON_ENDS):
                page_range.extend(range(0, ON_EACH_SIDE - 1))
                page_range.append(DOT)
                page_range.extend(range(
                    page_num - ON_EACH_SIDE, page_num + 1))
            else:
                page_range.extend(range(0, page_num + 1))
            if page_num < (paginator.num_pages - ON_EACH_SIDE - ON_ENDS - 1):
                page_range.extend(range(
                    page_num + 1, page_num + ON_EACH_SIDE + 1))
                page_range.append(DOT)
                page_range.extend(range(
                    paginator.num_pages - ON_ENDS, paginator.num_pages))
            else:
                page_range.extend(range(page_num + 1, paginator.num_pages))
    return {
        'page_range': page_range,
        'page_num': page_num,
        'results_var': context['results_var'],
        'query': context['query'],
    }


class IncludeEditorStuffNode(template.Node):

    def __init__(self, editor):
        self.editor = editor

    def render(self, context):
        editor = self.editor.resolve(context)
        if not isinstance(editor, basestring) or not re.match(r'^\w+$', editor):
            return ''
        template_name = 'filebrowser/editors/%s.html' % editor
        try:
            return render_to_string(template_name, context_instance=context)
        except template.TemplateDoesNotExist:
            return ''


@register.tag
def include_editor_stuff(parser, token):
    bits = token.split_contents()
    if len(bits) != 2:
        raise template.TemplateSyntaxError(
            '%r tag takes only one argument: the editor identifier.' % bits[0]
        )
    editor = parser.compile_filter(bits[1])
    return IncludeEditorStuffNode(editor)



import os
import operator
import re
import lemon

from time import gmtime, strftime, localtime, time

from django.contrib import messages
from django.core.exceptions import ImproperlyConfigured
from django.core.paginator import Paginator, InvalidPage
from django.core.urlresolvers import reverse
from django.forms.util import ErrorList
from django.http import Http404, HttpResponseRedirect
from django.shortcuts import render_to_response
from django.template.context import RequestContext
from django.utils.translation import ugettext as _
from django.views.i18n import javascript_catalog

from . import settings
from .base import FileObject
from .utils import query_helper


class FileBrowserAdmin(lemon.AppAdmin):

    upload_to = 'filebrowser'
    extensions = {'Folder': [''],
                  'Image': ['.jpg', '.jpeg', '.gif', '.png', '.tif', '.tiff'],
                  'Video': ['.mov', '.wmv', '.mpeg', '.mpg', '.avi', '.rm'],
                  'Document': ['.pdf', '.doc', '.rtf', '.txt', '.xls',
                               '.csv'],
                  'Audio': ['.mp3', '.mp4', '.wav', '.aiff', '.midi', '.m4p'],
                  'Code': ['.html', '.py', '.js', '.css']}
    select_formats = {'File': ['Folder', 'Document'],
                      'Image': ['Image'],
                      'Media': ['Video', 'Sound'],
                      'Document': ['Document']}
    order_by = '-date'
    list_per_page = 100
    convert_filenames = True
    max_upload_size = 10 * 1024 * 1024

    def __init__(self, *args, **kwargs):
        super(FileBrowserAdmin, self).__init__(*args, **kwargs)
        self.sort_order = 'desc' if self.order_by.startswith('-') else 'asc'
        self.sort_by = self.order_by.strip('+-')

    def get_urls(self):
        from django.conf.urls.defaults import patterns, url
        wrap = self.admin_site.admin_view
        return patterns(
            '',
            url(r'^$', wrap(self.browse_view), name='browse'),
            url(r'^mkdir/$', wrap(self.mkdir_view), name='mkdir'),
            url(r'^upload/$', wrap(self.upload_view), name='upload'),
            url(r'^rename/$', wrap(self.rename_view), name='rename'),
            url(r'^delete/$', wrap(self.delete_view), name='delete'),
            url(r'^jsi18n/$', wrap(self.i18n_javascript), name='jsi18n'),
        )

    @property
    def urls(self):
        return self.get_urls(), self.app_name, self.app_name

    def get_path(self, path):
        fullpath = os.path.join(settings.MEDIA_ROOT, self.upload_to, path)
        if path.startswith('.') or os.path.isabs(path) or \
           not os.path.isdir(fullpath):
            return None
        return path

    def get_file(self, path, filename):
        abs_path = os.path.join(
            settings.MEDIA_ROOT, self.upload_to, path, filename)
        if not os.path.isfile(abs_path) and not os.path.isdir(abs_path):
            return None
        return filename

    def get_breadcrumbs(self, query, path):
        breadcrumbs = []
        dir_query = ''
        if path:
            for item in path.split(os.sep):
                dir_query = os.path.join(dir_query, item)
                breadcrumbs.append([item, dir_query])
        return breadcrumbs

    def get_filterdate(self, filter_date, date_time):
        date_year = strftime('%Y', gmtime(date_time))
        date_month = strftime('%m', gmtime(date_time))
        date_day = strftime('%d', gmtime(date_time))
        if filter_date == 'today' and \
           int(date_year) == int(localtime()[0]) and \
           int(date_month) == int(localtime()[1]) and \
           int(date_day) == int(localtime()[2]):
            return 'true'
        if filter_date == 'thismonth' and date_time >= time() - 2592000:
            return 'true'
        if filter_date == 'thisyear' and \
           int(date_year) == int(localtime()[0]):
            return 'true'
        if filter_date == 'past7days' and date_time >= time() - 604800:
            return 'true'
        if filter_date == '':
            return 'true'
        return ''


    def sort_by_attr(self, seq, attr):
        intermed = map(None, map(getattr, seq, (attr,)*len(seq)),
                       xrange(len(seq)), seq)
        intermed.sort()
        return map(operator.getitem, intermed, (-1,) * len(intermed))

    def browse_view(self, request):
        query = request.GET.copy()

        directory = self.get_path('')
        if directory is None:
            try:
                os.makedirs(
                    os.path.join(settings.MEDIA_ROOT, self.upload_to), 0755)
            except OSError:
                raise ImproperlyConfigured(
                    "Error finding upload-folder. It seems that we can't "
                    " create it.")
            else:
                directory = self.get_path('')

        path = self.get_path(query.get('dir', ''))
        if path is None:
            messages.error(request, _('The requested folder does not exist.'))
            redirect_url = reverse('admin:filebrowser:browse')
            redirect_url += query_helper(query, 'dir')
            return HttpResponseRedirect(redirect_url)
        abs_path = os.path.join(settings.MEDIA_ROOT, self.upload_to, path)

        results_var = {'results_total': 0, 'delete_total': False}
        counter = {}
        for key in self.extensions.iterkeys():
            counter[key] = 0
        dir_list = os.listdir(abs_path)
        file_objects = []
        for filename in dir_list:
            if filename.startswith('.'):
                continue
            relative_path = os.path.join(self.upload_to, path, filename)
            file_object = FileObject(relative_path, self)
            q = request.GET.get('q')
            filter_type = request.GET.get('filter_type', file_object.filetype)
            filter_date = request.GET.get('filter_date', '')
            filter_date = self.get_filterdate(filter_date, file_object.date)
            append = file_object.filetype == filter_type and filter_date
            if q and not re.compile(q.lower(), re.M).search(filename.lower()):
                append = False
            if append:
                try:
                    if file_object.filetype != 'Folder' or \
                       file_object.is_empty:
                        results_var['delete_total'] = True
                except OSError:
                    continue
                else:
                    file_objects.append(file_object)
                    results_var['results_total'] += 1
            if file_object.filetype:
                counter[file_object.filetype] += 1
        query['o'] = request.GET.get('o', self.sort_by)
        query['ot'] = request.GET.get('ot', self.sort_order)
        file_objects = self.sort_by_attr(file_objects, query['o'])
        if query['ot'] == 'desc':
            file_objects.reverse()

        paginator = Paginator(file_objects, self.list_per_page)
        page_number = request.GET.get('p', 1)
        try:
            page_number = int(page_number)
        except ValueError:
            raise Http404
        try:
            page = paginator.page(page_number)
        except InvalidPage:
            raise Http404

        return render_to_response('filebrowser/browse.html',
            {'dir': path,
             'page': page,
             'results_var': results_var,
             'query': query,
             'title': _(u'File Browser'),
             'breadcrumbs': self.get_breadcrumbs(query, path),
             'breadcrumbs_title': '',
             'is_popup': request.GET.get('pop', False),
             'editor': request.GET.get('editor', None)},
            context_instance=RequestContext(request, current_app=self.admin_site.name))

    def mkdir_view(self, request):
        from .forms import MakeDirForm

        query = request.GET.copy()
        path = self.get_path(query.get('dir', ''))
        if path is None:
            messages.error(request, _('The requested folder does not exist.'))
            return HttpResponseRedirect(reverse('admin:filebrowser:browse'))
        abs_path = os.path.join(settings.MEDIA_ROOT, self.upload_to, path)

        if request.method == 'POST':
            form = MakeDirForm(abs_path, request.POST)
            if form.is_valid():
                server_path = os.path.join(abs_path,
                                           form.cleaned_data['dir_name'])
                try:
                    os.mkdir(server_path)
                    os.chmod(server_path, 0775)
                except OSError, (errno, strerror):
                    if errno == 13:
                        form.errors['dir_name'] = ErrorList(
                            [_('Permission denied.')]
                        )
                    else:
                        form.errors['dir_name'] = ErrorList(
                            [_('Error creating folder.')]
                        )
                else:
                    msg = _('The folder "%s" was successfully created.')
                    msg %= form.cleaned_data['dir_name']
                    messages.success(request, msg)
                    # on redirect, sort by date desc to see the new directory
                    # on top of the list
                    # remove filter in order to actually _see_ the new folder
                    # remove pagination
                    redirect_url = reverse("admin:filebrowser:browse")
                    redirect_url += query_helper(
                        query, 'o', 'q', 'p', 'ot', 'filter_type',
                        'filter_date', o='date', ot='desc')
                    return HttpResponseRedirect(redirect_url)
        else:
            form = MakeDirForm(abs_path)

        return render_to_response(
            'filebrowser/makedir.html',
            {'form': form,
             'query': query,
             'title': _(u'New Folder'),
             'breadcrumbs': self.get_breadcrumbs(query, path),
             'breadcrumbs_title': _(u'New Folder'),
             'is_popup': request.GET.get('pop', False)},
            context_instance=RequestContext(request))

    def upload_view(self, request):
        from .forms import FileUploadForm

        query = request.GET.copy()
        path = self.get_path(query.get('dir', ''))
        if path is None:
            messages.error(request, _('The requested folder does not exist.'))
            return HttpResponseRedirect(reverse('admin:filebrowser:browse'))
        abs_path = os.path.join(settings.MEDIA_ROOT, self.upload_to, path)

        if request.method == 'POST':
            form = FileUploadForm(abs_path, request.POST, request.FILES)
            if form.is_valid():
                uploaded_file = form.cleaned_data['file']
                f = open(os.path.join(abs_path, uploaded_file.name), 'wb+')
                for chunk in uploaded_file.chunks():
                    f.write(chunk)
                f.close()
                redirect_url = reverse("admin:filebrowser:browse")
                redirect_url += query_helper(
                    query, 'o', 'q', 'p', 'ot', 'filter_type', 'filter_date',
                    ot='desc', o='date')
                return HttpResponseRedirect(redirect_url)
        else:
            form = FileUploadForm(abs_path)

        return render_to_response(
            'filebrowser/upload.html',
            {'form': form,
             'query': query,
             'title': _(u'Upload File'),
             'is_popup': request.GET.get('pop', False)},
            context_instance=RequestContext(request))

    def rename_view(self, request):
        from .forms import RenameForm

        query = request.GET.copy()
        path = self.get_path(query.get('dir', ''))
        filename = self.get_file(
            query.get('dir', ''), query.get('filename', ''))
        if path is None or filename is None:
            if path is None:
                msg = _('The requested folder does not exist.')
            else:
                msg = _('The requested file does not exist.')
            messages.error(request, msg)
            return HttpResponseRedirect(reverse('admin:filebrowser:browse'))
        abs_path = os.path.join(settings.MEDIA_ROOT, self.upload_to, path)
        file_extension = os.path.splitext(filename)[1].lower()

        if request.method == 'POST':
            form = RenameForm(file_extension, abs_path, request.POST)
            if form.is_valid():
                server_path = os.path.join(
                    settings.MEDIA_ROOT, self.upload_to, path, filename)
                new_filename = form.cleaned_data['name'] + file_extension
                new_server_path = os.path.join(
                    settings.MEDIA_ROOT, self.upload_to, path, new_filename)
                try:
                    os.rename(server_path, new_server_path)
                    if os.path.isdir(new_server_path):
                        msg = _('The folder was successfully renamed.')
                    else:
                        msg = _('The file was successfully renamed.')
                except OSError, (errno, strerror):
                    msg = _(u'OS error while delete.')
                else:
                    messages.success(request, msg)
                    redirect_url = reverse("admin:filebrowser:browse")
                    redirect_url += query_helper(query, 'filename')
                    return HttpResponseRedirect(redirect_url)
        else:
            form = RenameForm(file_extension, abs_path)

        return render_to_response('filebrowser/rename.html', {
            'form': form,
            'query': query,
            'file_extension': file_extension,
            'title': _(u'Rename "%s"') % filename,
            'breadcrumbs': self.get_breadcrumbs(query, path),
            'breadcrumbs_title': _(u'Rename'),
            'is_popup': request.GET.get('pop', False)
        }, context_instance=RequestContext(request))

    def delete_view(self, request):
        query = request.GET.copy()
        path = self.get_path(query.get('dir', ''))
        filename = self.get_file(
            query.get('dir', ''), query.get('filename', ''))
        if path is None or filename is None:
            if path is None:
                msg = _('The requested folder does not exist.')
            else:
                msg = _('The requested file does not exist.')
            messages.error(request, msg)
            return HttpResponseRedirect(reverse('admin:filebrowser:browse'))
        abs_path = os.path.join(settings.MEDIA_ROOT, self.upload_to, path)

        msg = ''
        if request.GET:
            if request.GET.get('filetype') != 'Folder':
                try:
                    os.unlink(os.path.join(abs_path, filename))
                except OSError:
                    msg = _(u'OS error while delete.')
                else:
                    msg = _('The file "%s" was successfully deleted.')
                    msg %= filename.lower()
                    messages.success(request, msg)
                    redirect_url = reverse('admin:filebrowser:browse')
                    redirect_url += query_helper(
                        query, 'filename', 'filetype')
                    return HttpResponseRedirect(redirect_url)
            else:
                try:
                    os.rmdir(os.path.join(abs_path, filename))
                except OSError:
                    msg = _(u'OS error while delete.')
                else:
                    msg = _('The folder "%s" was successfully deleted.')
                    msg %= filename.lower()
                    messages.success(request, msg)
                    redirect_url = reverse('admin:filebrowser:browse')
                    redirect_url += query_helper(
                        query, 'filename',  'filetype')
                    return HttpResponseRedirect(redirect_url)
        if msg:
            messages.error(request, msg)

        return render_to_response('filebrowser/browse.html', {
            'dir': dir_name,
            'file': request.GET.get('filename', ''),
            'query': query,
            'breadcrumbs': self.get_breadcrumbs(query, dir_name),
            'breadcrumbs_title': '',
            'is_popup': request.GET.get('pop', False)
        }, context_instance=RequestContext(request))

    def i18n_javascript(self, request):
        return javascript_catalog(request, packages='lemon.filebrowser')


lemon.site.register_app('filebrowser', FileBrowserAdmin)
